@@ -118,9 +118,6 @@ int pointerY     = 0;
 #define LOWER_WINDOW_BUTTON_DEFAULT    6
 #define LOWER_WINDOW_MODIFIERS_DEFAULT CompAltMask
 
-#define RAISE_WINDOW_BUTTON_DEFAULT    2
-#define RAISE_WINDOW_MODIFIERS_DEFAULT ControlMask
-
 #define SHOW_DESKTOP_KEY_DEFAULT       "d"
 #define SHOW_DESKTOP_MODIFIERS_DEFAULT (CompAltMask | ControlMask)
 
@@ -420,25 +417,6 @@ lower (CompDisplay     *d,
     return TRUE;
 }
 
-static Bool
-raise (CompDisplay     *d,
-       CompAction      *action,
-       CompActionState state,
-       CompOption      *option,
-       int	       nOption)
-{
-    CompWindow *w;
-    Window     xid;
-
-    xid = getIntOptionNamed (option, nOption, "window", 0);
-
-    w = findTopLevelWindowAtDisplay (d, xid);
-    if (w)
-	raiseWindow (w);
-
-    return TRUE;
-}
-
 static void
 changeWindowOpacity (CompWindow *w,
 		     int	direction)
@@ -985,21 +963,6 @@ compDisplayInitOptions (CompDisplay *display,
     o->value.action.button.modifiers = LOWER_WINDOW_MODIFIERS_DEFAULT;
     o->value.action.button.button    = LOWER_WINDOW_BUTTON_DEFAULT;
 
-    o = &display->opt[COMP_DISPLAY_OPTION_RAISE_WINDOW];
-    o->name			     = "raise_window";
-    o->shortDesc		     = N_("Raise Window");
-    o->longDesc			     = N_("Raise window above other windows");
-    o->type			     = CompOptionTypeAction;
-    o->value.action.initiate	     = raise;
-    o->value.action.terminate        = 0;
-    o->value.action.bell	     = FALSE;
-    o->value.action.edgeMask	     = 0;
-    o->value.action.state	     = CompActionStateInitKey;
-    o->value.action.state	    |= CompActionStateInitButton;
-    o->value.action.type	     = CompBindingTypeButton;
-    o->value.action.button.modifiers = RAISE_WINDOW_MODIFIERS_DEFAULT;
-    o->value.action.button.button    = RAISE_WINDOW_BUTTON_DEFAULT;
-
     o = &display->opt[COMP_DISPLAY_OPTION_OPACITY_INCREASE];
     o->name			     = "opacity_increase";
     o->shortDesc		     = N_("Increase Opacity");
@@ -1299,7 +1262,6 @@ setDisplayOption (CompDisplay     *display,
     case COMP_DISPLAY_OPTION_TOGGLE_WINDOW_MAXIMIZED_HORZ:
     case COMP_DISPLAY_OPTION_TOGGLE_WINDOW_MAXIMIZED_VERT:
     case COMP_DISPLAY_OPTION_TOGGLE_WINDOW_SHADED:
-    case COMP_DISPLAY_OPTION_RAISE_WINDOW:
 	if (setDisplayAction (display, o, value))
 	    return TRUE;
 	break;
@@ -1587,7 +1549,8 @@ getTimeToNextRedraw (CompScreen     *s,
     if (diff < 0)
 	diff = 0;
 
-    if (idle)
+    if (idle ||
+	(s->getVideoSync && s->opt[COMP_SCREEN_OPTION_SYNC_TO_VBLANK].value.b))
     {
 	if (s->timeMult > 1)
 	{
@@ -1651,25 +1614,22 @@ void
 updateModifierMappings (CompDisplay *d)
 {
     unsigned int    modMask[CompModNum];
-    int		    i, minKeycode = 0, maxKeycode = 0, keysymsPerKeycode = 0;
-    KeySym	    * keyboard_sym;	
+    int		    i, minKeycode, maxKeycode, keysymsPerKeycode = 0;
 
     for (i = 0; i < CompModNum; i++)
 	modMask[i] = 0;
 
     XDisplayKeycodes (d->display, &minKeycode, &maxKeycode);
-    keyboard_sym = XGetKeyboardMapping (d->display, minKeycode, (maxKeycode - minKeycode + 1), &keysymsPerKeycode);
+    XGetKeyboardMapping (d->display, minKeycode, (maxKeycode - minKeycode + 1),
+			 &keysymsPerKeycode);
 
-    if (keyboard_sym)
-	XFree(keyboard_sym);  
-	
     if (d->modMap)
 	XFreeModifiermap (d->modMap);
 
     d->modMap = XGetModifierMapping (d->display);
     if (d->modMap && d->modMap->max_keypermod > 0)
     {
-	KeySym keysym = 0;
+	KeySym keysym;
 	int    index, size, mask;
 
 	size = maskTableSize * d->modMap->max_keypermod;
@@ -1840,23 +1800,47 @@ handleTimeouts (struct timeval *tv)
     lastTimeout = *tv;
 }
 
+static void
+waitForVideoSync (CompScreen *s)
+{
+    unsigned int sync;
+
+    if (!s->opt[COMP_SCREEN_OPTION_SYNC_TO_VBLANK].value.b)
+	return;
+
+    /* we currently can't handle sync to vblank when we have more than one
+       output device */
+    if (s->nOutputDev > 1)
+	return;
+
+    if (s->getVideoSync)
+    {
+	glFlush ();
+
+	(*s->getVideoSync) (&sync);
+	(*s->waitVideoSync) (2, (sync + 1) % 2, &sync);
+    }
+}
+
 void
 eventLoop (void)
 {
     XEvent	   event;
-    int		   timeDiff;
+    int		   timeDiff, i;
     struct timeval tv;
-    Region	   tmpRegion;
+    Region	   tmpRegion, outputRegion;
     CompDisplay    *display = compDisplays;
     CompScreen	   *s;
     int		   time, timeToNextRedraw = 0;
     CompWindow	   *w;
-    unsigned int   damageMask;
+    unsigned int   damageMask, mask;
 
     tmpRegion = XCreateRegion ();
-    if (!tmpRegion)
+    outputRegion = XCreateRegion ();
+    if (!tmpRegion || !outputRegion)
     {
-	fprintf (stderr, "%s: Couldn't create region\n", programName);
+	fprintf (stderr, "%s: Couldn't create temporary regions\n",
+		 programName);
 	return;
     }
 
@@ -2019,9 +2003,9 @@ eventLoop (void)
 		    {
 			XIntersectRegion (s->damage, &s->region, tmpRegion);
 
-			if (tmpRegion->numRects  == 1	 &&
-			    tmpRegion->rects->x1 == 0	 &&
-			    tmpRegion->rects->y1 == 0	 &&
+			if (tmpRegion->numRects  == 1	     &&
+			    tmpRegion->rects->x1 == 0	     &&
+			    tmpRegion->rects->y1 == 0	     &&
 			    tmpRegion->rects->x2 == s->width &&
 			    tmpRegion->rects->y2 == s->height)
 			    damageScreen (s);
@@ -2029,91 +2013,118 @@ eventLoop (void)
 
 		    EMPTY_REGION (s->damage);
 
-		    if (s->damageMask & COMP_SCREEN_DAMAGE_ALL_MASK)
+		    mask = s->damageMask;
+		    s->damageMask = 0;
+
+		    for (i = 0; i < s->nOutputDev; i++)
 		    {
-			s->damageMask = 0;
+			if (s->nOutputDev > 1)
+			    glScissor (s->outputDev[i].region.extents.x1,
+				       s->outputDev[i].region.extents.y1,
+				       s->outputDev[i].region.extents.x2 -
+				       s->outputDev[i].region.extents.x1,
+				       s->outputDev[i].region.extents.y2 -
+				       s->outputDev[i].region.extents.y1);
 
-			(*s->paintScreen) (s,
-					   &defaultScreenPaintAttrib,
-					   &s->region,
-					   PAINT_SCREEN_REGION_MASK |
-					   PAINT_SCREEN_FULL_MASK);
-
-			glXSwapBuffers (s->display->display, s->root);
-		    }
-		    else if (s->damageMask & COMP_SCREEN_DAMAGE_REGION_MASK)
-		    {
-			s->damageMask = 0;
-
-			if ((*s->paintScreen) (s,
-					       &defaultScreenPaintAttrib,
-					       tmpRegion,
-					       PAINT_SCREEN_REGION_MASK))
+			if (mask & COMP_SCREEN_DAMAGE_ALL_MASK)
 			{
-			    BoxPtr pBox;
-			    int    nBox, y;
+			    (*s->paintScreen) (s,
+					       &defaultScreenPaintAttrib,
+					       &s->outputDev[i].region, i,
+					       PAINT_SCREEN_REGION_MASK |
+					       PAINT_SCREEN_FULL_MASK);
 
-			    pBox = tmpRegion->rects;
-			    nBox = tmpRegion->numRects;
-
-			    if (s->copySubBuffer)
+			    if (i + 1 == s->nOutputDev)
 			    {
-				while (nBox--)
+				waitForVideoSync (s);
+				glXSwapBuffers (s->display->display, s->output);
+			    }
+			}
+			else if (mask & COMP_SCREEN_DAMAGE_REGION_MASK)
+			{
+			    XIntersectRegion (tmpRegion,
+					      &s->outputDev[i].region,
+					      outputRegion);
+
+			    if ((*s->paintScreen) (s,
+						   &defaultScreenPaintAttrib,
+						   outputRegion, i,
+						   PAINT_SCREEN_REGION_MASK))
+			    {
+				BoxPtr pBox;
+				int    nBox, y;
+
+				pBox = outputRegion->rects;
+				nBox = outputRegion->numRects;
+
+				waitForVideoSync (s);
+
+				if (s->copySubBuffer)
 				{
-				    y = s->height - pBox->y2;
+				    while (nBox--)
+				    {
+					y = s->height - pBox->y2;
 
-				    (*s->copySubBuffer) (s->display->display,
-							 s->root,
-							 pBox->x1, y,
-							 pBox->x2 - pBox->x1,
-							 pBox->y2 - pBox->y1);
+					(*s->copySubBuffer) (display->display,
+							     s->output,
+							     pBox->x1, y,
+							     pBox->x2 -
+							     pBox->x1,
+							     pBox->y2 -
+							     pBox->y1);
 
-				    pBox++;
+					pBox++;
+				    }
+				}
+				else
+				{
+				    glEnable (GL_SCISSOR_TEST);
+				    glDrawBuffer (GL_FRONT);
+
+				    while (nBox--)
+				    {
+					y = s->height - pBox->y2;
+
+					glBitmap (0, 0, 0, 0,
+						  pBox->x1 - s->rasterX,
+						  y - s->rasterY,
+						  NULL);
+
+					s->rasterX = pBox->x1;
+					s->rasterY = y;
+
+					glScissor (pBox->x1, y,
+						   pBox->x2 - pBox->x1,
+						   pBox->y2 - pBox->y1);
+
+					glCopyPixels (pBox->x1, y,
+						      pBox->x2 - pBox->x1,
+						      pBox->y2 - pBox->y1,
+						      GL_COLOR);
+
+					pBox++;
+				    }
+
+				    glDrawBuffer (GL_BACK);
+				    glDisable (GL_SCISSOR_TEST);
+				    glFlush ();
 				}
 			    }
 			    else
 			    {
-				glEnable (GL_SCISSOR_TEST);
-				glDrawBuffer (GL_FRONT);
+				(*s->paintScreen) (s,
+						   &defaultScreenPaintAttrib,
+						   &s->outputDev[i].region, i,
+						   PAINT_SCREEN_FULL_MASK);
 
-				while (nBox--)
+				if (i + 1 == s->nOutputDev)
 				{
-				    y = s->height - pBox->y2;
-
-				    glBitmap (0, 0, 0, 0,
-					      pBox->x1 - s->rasterX,
-					      y - s->rasterY,
-					      NULL);
-
-				    s->rasterX = pBox->x1;
-				    s->rasterY = y;
-
-				    glScissor (pBox->x1, y,
-					       pBox->x2 - pBox->x1,
-					       pBox->y2 - pBox->y1);
-
-				    glCopyPixels (pBox->x1, y,
-						  pBox->x2 - pBox->x1,
-						  pBox->y2 - pBox->y1,
-						  GL_COLOR);
-
-				    pBox++;
+				    waitForVideoSync (s);
+				    glXSwapBuffers (display->display,
+						    s->output);
 				}
-
-				glDrawBuffer (GL_BACK);
-				glDisable (GL_SCISSOR_TEST);
-				glFlush ();
 			    }
 			}
-			else
-			{
-			    (*s->paintScreen) (s,
-					       &defaultScreenPaintAttrib,
-					       &s->region,
-					       PAINT_SCREEN_FULL_MASK);
-
-			    glXSwapBuffers (s->display->display, s->root);
-			}
 		    }
 
 		    s->lastRedraw = tv;
@@ -2341,7 +2352,6 @@ addScreenActions (CompDisplay *d, CompScreen *s)
     addScreenAction (s,
 		     &d->opt[COMP_DISPLAY_OPTION_SLOW_ANIMATIONS].value.action);
     addScreenAction (s, &d->opt[COMP_DISPLAY_OPTION_LOWER_WINDOW].value.action);
-    addScreenAction (s, &d->opt[COMP_DISPLAY_OPTION_RAISE_WINDOW].value.action);
     addScreenAction (s,
 		     &d->opt[COMP_DISPLAY_OPTION_OPACITY_INCREASE].value.action);
     addScreenAction (s,
@@ -2669,28 +2679,24 @@ addDisplay (char *name,
 	d->xkbEvent = d->xkbError = -1;
     }
 
+    d->xineramaExtension = XineramaQueryExtension (dpy,
+						   &d->xineramaEvent,
+						   &d->xineramaError);
+    if (d->xineramaExtension)
+    {
+	d->screenInfo = XineramaQueryScreens (dpy, &d->nScreenInfo);
+    }
+    else
+    {
+	d->screenInfo  = NULL;
+	d->nScreenInfo = 0;
+    }
+
     compDisplays = d;
 
     d->escapeKeyCode = XKeysymToKeycode (dpy, XStringToKeysym ("Escape"));
     d->returnKeyCode = XKeysymToKeycode (dpy, XStringToKeysym ("Return"));
 
-    {
-	XSetWindowAttributes attr;
-	Atom atom;
-
-	attr.override_redirect = TRUE;
-	d->glIncludeInferiorsOwner =
-	    XCreateWindow (dpy, XRootWindow (dpy, 0),
-			   -100, -100, 1, 1, 0,
-			   CopyFromParent, CopyFromParent,
-			   CopyFromParent,
-			   CWOverrideRedirect,
-			   &attr);
-
-	atom = XInternAtom (dpy, "_COMPIZ_GL_INCLUDE_INFERIORS", False);
-	XSetSelectionOwner(dpy, atom, d->glIncludeInferiorsOwner, CurrentTime);
-    }
-
     for (i = 0; i < ScreenCount (dpy); i++)
     {
 	Window		     newWmSnOwner = None, newCmSnOwner = None;
