@@ -318,12 +318,11 @@ triggerKeyPressBindings (CompDisplay *d,
 			 CompOption  *argument,
 			 int	     nArgument)
 {
-    CompActionState state = CompActionStateInitKey;
+    CompActionState state = 0;
     CompAction	    *action;
     unsigned int    modMask = REAL_MOD_MASK & ~d->ignoredModMask;
     unsigned int    bindMods;
 
-    state = 0;
     if (event->xkey.keycode == d->escapeKeyCode)
 	state = CompActionStateCancel;
     else if (event->xkey.keycode == d->returnKeyCode)
@@ -331,19 +330,23 @@ triggerKeyPressBindings (CompDisplay *d,
 
     if (state)
     {
-	while (nOption--)
+	CompOption *o = option;
+	int	   n = nOption;
+
+	while (n--)
 	{
-	    if (option->type == CompOptionTypeAction)
+	    if (o->type == CompOptionTypeAction)
 	    {
-		if (option->value.action.terminate)
-		    (*option->value.action.terminate) (d, &option->value.action,
-						       state, NULL, 0);
+		if (o->value.action.terminate)
+		    (*o->value.action.terminate) (d, &o->value.action,
+						  state, NULL, 0);
 	    }
 
-	    option++;
+	    o++;
 	}
 
-	return FALSE;
+	if (state == CompActionStateCancel)
+	    return FALSE;
     }
 
     state = CompActionStateInitKey;
@@ -1063,6 +1066,22 @@ handleEvent (CompDisplay *d,
     CompScreen *s;
     CompWindow *w;
 
+    switch (event->type) {
+    case ButtonPress:
+	s = findScreenAtDisplay (d, event->xbutton.root);
+	if (s)
+	    setCurrentOutput (s, outputDeviceForPoint (s,
+						       event->xbutton.x_root,
+						       event->xbutton.y_root));
+	break;
+    case KeyPress:
+	w = findWindowAtDisplay (d, d->activeWindow);
+	if (w)
+	    setCurrentOutput (w->screen, outputDeviceForWindow (w));
+    default:
+	break;
+    }
+
     if (handleActionEvent (d, event))
     {
 	if (!d->screens->maxGrab)
@@ -1145,7 +1164,12 @@ handleEvent (CompDisplay *d,
 	s = findScreenAtDisplay (d, event->xcreatewindow.parent);
 	if (s)
 	{
-	    addWindow (s, event->xcreatewindow.window, getTopWindow (s));
+	    /* The first time some client asks for the composite
+	     * overlay window, the X server creates it, which causes
+	     * an errorneous CreateNotify event.  We catch it and
+	     * ignore it. */
+	    if (s->overlay != event->xcreatewindow.window)
+		addWindow (s, event->xcreatewindow.window, getTopWindow (s));
 	}
 	break;
     case DestroyNotify:
@@ -1249,6 +1273,7 @@ handleEvent (CompDisplay *d,
 			moveInputFocusToWindow (w);
 		}
 	    }
+
 	    if (!s->maxGrab)
 		XAllowEvents (d->display, ReplayPointer, event->xbutton.time);
 	}
@@ -1271,7 +1296,6 @@ handleEvent (CompDisplay *d,
 			w->activeNum = s->activeNum++;
 		}
 	    }
-
 	}
 	else if (event->xproperty.atom == d->winTypeAtom)
 	{
@@ -1758,7 +1782,8 @@ handleEvent (CompDisplay *d,
 
 	    if (!(w->state & CompWindowStateHiddenMask))
 	    {
-		w->initialViewport = w->screen->x;
+		w->initialViewportX = w->screen->x;
+		w->initialViewportY = w->screen->y;
 
 		applyStartupProperties (w->screen, w);
 
